@@ -1,15 +1,16 @@
-# -*- coding: utf-8 -*-
 """
 This module handles starting the maestral daemon on user login and supports multiple
 platform specific backends such as launchd or systemd.
 
 Note that launchd agents will not show as "login items" in macOS system preferences. As
 a result, the user does not have a convenient UI to remove Maestral autostart entries
-manually outside of Maestral itself. Login items however only support app bundles and
+manually outside Maestral itself. Login items however only support app bundles and
 provide no option to pass command line arguments to the app. They would therefore
-neither support pip installed packages or multiple configurations.
+neither support pip installed packages nor multiple configurations.
 """
 
+from __future__ import annotations
+
 # system imports
 import os
 import os.path as osp
@@ -23,16 +24,18 @@ import plistlib
 import configparser
 from pathlib import Path
 from enum import Enum
-from typing import Optional, Dict, Any
+from typing import Any
 
 try:
-    from importlib.metadata import files, PackageNotFoundError  # type: ignore
+    from importlib.metadata import files, PackageNotFoundError
 except ImportError:  # Python 3.7 and lower
-    from importlib_metadata import files, PackageNotFoundError  # type: ignore
+    from importlib_metadata import files, PackageNotFoundError
 
 # local imports
 from .utils.appdirs import get_home_dir, get_conf_path, get_data_path
+from .utils.integration import cat
 from .constants import BUNDLE_ID, ENV
+from .exceptions import MaestralApiError
 
 
 class SupportedImplementations(Enum):
@@ -68,7 +71,7 @@ class AutoStartSystemd(AutoStartBase):
     :param unit_dict: Dictionary of additional keys and values for the Unit section.
     :param service_dict: Dictionary of additional keys and values for the Service
         section.
-    :param install_dict: Dictionary of additional keys and values for the Install
+    :param install_dict: Dictionary of additional keys and values for the "Install"
         section.
     """
 
@@ -76,9 +79,9 @@ class AutoStartSystemd(AutoStartBase):
         self,
         service_name: str,
         start_cmd: str,
-        unit_dict: Optional[Dict[str, str]] = None,
-        service_dict: Optional[Dict[str, str]] = None,
-        install_dict: Optional[Dict[str, str]] = None,
+        unit_dict: dict[str, str] | None = None,
+        service_dict: dict[str, str] | None = None,
+        install_dict: dict[str, str] | None = None,
     ) -> None:
         super().__init__()
 
@@ -114,10 +117,16 @@ class AutoStartSystemd(AutoStartBase):
             self.service_config.write(f)
 
     def enable(self) -> None:
-        subprocess.check_output(["systemctl", "--user", "enable", self.service_name])
+        res = subprocess.run(["systemctl", "--user", "enable", self.service_name])
+
+        if res.returncode != 0:
+            raise MaestralApiError("Could not enable autostart", str(res.stderr))
 
     def disable(self) -> None:
-        subprocess.check_output(["systemctl", "--user", "disable", self.service_name])
+        res = subprocess.run(["systemctl", "--user", "disable", self.service_name])
+
+        if res.returncode != 0:
+            raise MaestralApiError("Could not disable autostart", str(res.stderr))
 
     @property
     def enabled(self) -> bool:
@@ -131,26 +140,26 @@ class AutoStartSystemd(AutoStartBase):
 class AutoStartLaunchd(AutoStartBase):
     """Autostart backend for launchd
 
-    :param bundle_id: Bundle ID for the, e.g., "com.google.calendar".
+    :param launchd_id: Identifier for the launchd job, e.g., "com.google.calendar".
     :param start_cmd: Absolute path to executable and optional program arguments.
     :param kwargs: Additional key, value pairs to add to plist. Values may be strings,
         booleans, lists or dictionaries.
     """
 
-    def __init__(self, bundle_id: str, start_cmd: str, **kwargs) -> None:
-
+    def __init__(self, launchd_id: str, start_cmd: str, **kwargs: Any) -> None:
         super().__init__()
-        filename = bundle_id + ".plist"
+        filename = launchd_id + ".plist"
 
         self.path = osp.join(get_home_dir(), "Library", "LaunchAgents")
         self.destination = osp.join(self.path, filename)
 
-        self.plist_dict: Dict[str, Any] = dict(
-            Label=None, ProcessType="Interactive", ProgramArguments=[], RunAtLoad=True
-        )
+        self.plist_dict = {
+            "Label": str(launchd_id),
+            "ProcessType": "Interactive",
+            "RunAtLoad": True,
+            "ProgramArguments": shlex.split(start_cmd),
+        }
 
-        self.plist_dict["Label"] = str(bundle_id)
-        self.plist_dict["ProgramArguments"] = shlex.split(start_cmd)
         self.plist_dict.update(kwargs)
 
     def enable(self) -> None:
@@ -190,7 +199,7 @@ class AutoStartXDGDesktop(AutoStartBase):
     """
 
     def __init__(
-        self, app_name: str, start_cmd: str, filename: Optional[str], **kwargs: str
+        self, app_name: str, start_cmd: str, filename: str | None, **kwargs: str
     ) -> None:
         super().__init__()
 
@@ -211,18 +220,23 @@ class AutoStartXDGDesktop(AutoStartBase):
         self.destination = get_conf_path("autostart", filename)
 
     def enable(self) -> None:
+        try:
+            with open(self.destination, "w") as f:
+                self.config.write(f, space_around_delimiters=False)
 
-        with open(self.destination, "w") as f:
-            self.config.write(f, space_around_delimiters=False)
+            st = os.stat(self.destination)
+            os.chmod(self.destination, st.st_mode | stat.S_IEXEC)
 
-        st = os.stat(self.destination)
-        os.chmod(self.destination, st.st_mode | stat.S_IEXEC)
+        except Exception as exc:
+            raise MaestralApiError("Could not enable autostart", exc.args[0])
 
     def disable(self) -> None:
         try:
             os.unlink(self.destination)
-        except FileNotFoundError:
+        except (FileNotFoundError, NotADirectoryError):
             pass
+        except Exception as exc:
+            raise MaestralApiError("Could not enable autostart", exc.args[0])
 
     @property
     def enabled(self) -> bool:
@@ -230,20 +244,16 @@ class AutoStartXDGDesktop(AutoStartBase):
         return os.path.isfile(self.destination)
 
 
-def get_available_implementation() -> Optional[SupportedImplementations]:
+def get_available_implementation() -> SupportedImplementations | None:
     """Returns the supported implementation depending on the platform."""
-
     system = platform.system()
-
     if system == "Darwin":
         return SupportedImplementations.launchd
-    else:
-        try:
-            res = subprocess.check_output(["ps", "-p", "1"]).decode()
-        except (FileNotFoundError, subprocess.CalledProcessError):
-            return None
-        else:
-            return SupportedImplementations.systemd if "systemd" in res else None
+    if system == "Linux":
+        init_command = cat(Path("/proc/1/comm"))
+        if init_command is not None and b"systemd" in init_command:
+            return SupportedImplementations.systemd
+    return None
 
 
 def get_maestral_command_path() -> str:
@@ -251,14 +261,13 @@ def get_maestral_command_path() -> str:
     Returns the path to the maestral executable. May be an empty string if the
     executable cannot be found.
     """
-
     try:
         dist_files = files("maestral")
     except PackageNotFoundError:
         # we may have had installation issues
         dist_files = []
 
-    path: Optional[os.PathLike]
+    path: os.PathLike[str] | None
 
     if dist_files:
         try:
@@ -292,22 +301,21 @@ class AutoStart:
     _impl: AutoStartBase
 
     def __init__(self, config_name: str) -> None:
-
         self.maestral_path = get_maestral_command_path()
         self.implementation = get_available_implementation()
 
         start_cmd = f"{self.maestral_path} start -f -c {config_name}"
-        bundle_id = f"{BUNDLE_ID}-daemon.{config_name}"
+        launchd_id = f"{BUNDLE_ID}-daemon.{config_name}"
 
         if self.implementation == SupportedImplementations.launchd:
             self._impl = AutoStartLaunchd(
-                bundle_id,
+                launchd_id,
                 start_cmd,
                 EnvironmentVariables=ENV,
+                AssociatedBundleIdentifiers=BUNDLE_ID,
             )
 
         elif self.implementation == SupportedImplementations.systemd:
-
             notify_failure = (
                 "if [ ${SERVICE_RESULT} != success ]; "
                 "then notify-send Maestral 'Daemon failed: ${SERVICE_RESULT}'; "
@@ -349,7 +357,6 @@ class AutoStart:
 
     def enable(self) -> None:
         """Enable autostart."""
-
         if self.enabled:
             return
 
@@ -360,7 +367,6 @@ class AutoStart:
 
     def disable(self) -> None:
         """Disable autostart."""
-
         if not self.enabled:
             return
 
