@@ -18,9 +18,10 @@ def _style_hint(hint: str) -> str:
 
 
 def prompt(
-    message: str, default: str | None = None, validate: Callable | None = None
+    message: str,
+    default: str | None = None,
+    validate: Callable[[str], bool] | None = None,
 ) -> str:
-
     import survey
 
     styled_message = _style_message(message)
@@ -31,13 +32,10 @@ def prompt(
         else:
             return True
 
-    res = survey.input(styled_message, default=default, check=check)
-
-    return res
+    return survey.input(styled_message, default=default, check=check)
 
 
 def confirm(message: str, default: bool | None = True) -> bool:
-
     import survey
 
     styled_message = _style_message(message)
@@ -45,8 +43,7 @@ def confirm(message: str, default: bool | None = True) -> bool:
     return survey.confirm(styled_message, default=default)
 
 
-def select(message: str, options: Sequence[str], hint="") -> int:
-
+def select(message: str, options: Sequence[str], hint: str = "") -> int:
     import survey
 
     try:
@@ -61,8 +58,7 @@ def select(message: str, options: Sequence[str], hint="") -> int:
         raise
 
 
-def select_multiple(message: str, options: Sequence[str], hint="") -> list[int]:
-
+def select_multiple(message: str, options: Sequence[str], hint: str = "") -> list[int]:
     import survey
 
     try:
@@ -93,14 +89,12 @@ def select_multiple(message: str, options: Sequence[str], hint="") -> list[int]:
 def select_path(
     message: str,
     default: str | None = None,
-    validate: Callable = lambda x: True,
+    validate: Callable[[str], bool] = lambda x: True,
     exists: bool = False,
     files_allowed: bool = True,
     dirs_allowed: bool = True,
 ) -> str:
-
     import os
-
     import survey
     import wrapio
 
@@ -111,7 +105,6 @@ def select_path(
     failed = False
 
     def check(value: str) -> bool:
-
         nonlocal failed
 
         if value == "" and default:
@@ -138,11 +131,9 @@ def select_path(
 
         return not failed
 
-    res = survey.input(
+    return survey.input(
         styled_message,
         default=default,
         callback=track.invoke,
         check=check,
     )
-
-    return res
