@@ -18,15 +18,15 @@ import copy
 import logging
 import configparser as cp
 from threading import RLock
-from collections import abc
-from typing import Iterator, Any, Dict
+from typing import Iterator, Any, Dict, TypeVar, MutableSet
 
 from packaging.version import Version
 
 
 logger = logging.getLogger(__name__)
 
-DefaultsType = Dict[str, Dict[str, Any]]
+_DefaultsType = Dict[str, Dict[str, Any]]
+_T = TypeVar("_T")
 
 # =============================================================================
 # Auxiliary classes
@@ -69,7 +69,6 @@ class DefaultsConfig(cp.ConfigParser):
 
     def save(self) -> None:
         """Save config into the associated file."""
-
         os.makedirs(self._dirname, exist_ok=True)
 
         with open(self.config_path, "w", encoding="utf-8") as configfile:
@@ -107,7 +106,7 @@ class UserConfig(DefaultsConfig):
     def __init__(
         self,
         path: str,
-        defaults: DefaultsType | None = None,
+        defaults: _DefaultsType | None = None,
         load: bool = True,
         version: Version = Version("0.0.0"),
         backup: bool = False,
@@ -145,7 +144,6 @@ class UserConfig(DefaultsConfig):
             # Updating defaults only if major/minor version is different.
 
             if version != old_version:
-
                 if backup:
                     self._make_backup(old_version)
 
@@ -164,15 +162,14 @@ class UserConfig(DefaultsConfig):
     # --- Helpers and checkers ---------------------------------------------------------
 
     def _set_defaults(
-        self, version: Version, defaults: DefaultsType | None
-    ) -> DefaultsType:
+        self, version: Version, defaults: _DefaultsType | None
+    ) -> _DefaultsType:
         """
         Check if defaults are valid and update defaults values.
 
         :param version: The config version.
         :param defaults: New default config values.
         """
-
         if defaults:
             defaults = copy.deepcopy(defaults)
         else:
@@ -193,7 +190,6 @@ class UserConfig(DefaultsConfig):
         :param version: If a version is provided, it will be appended to the backup
             file name.
         """
-
         os.makedirs(self._dirname, exist_ok=True)
         backup_path = self.backup_path_for_version(version)
 
@@ -209,7 +205,6 @@ class UserConfig(DefaultsConfig):
 
         :param path: Path of config file to load.
         """
-
         with self._lock:
             try:
                 self.read(path, encoding="utf-8")
@@ -279,7 +274,6 @@ class UserConfig(DefaultsConfig):
         :param version: New version to set.
         :param save: Whether to save changes to drive.
         """
-
         with self._lock:
             self.set(
                 UserConfig.DEFAULT_SECTION_NAME, "version", str(version), save=save
@@ -296,15 +290,12 @@ class UserConfig(DefaultsConfig):
         :param section: The section to reset. If not given, reset all sections.
         :param save: Whether to save the changes to the drive.
         """
-
         with self._lock:
-
             for sec, options in self.default_config.items():
                 if section is None or section == sec:
                     for option in options:
                         value = options[option]
                         self._set(sec, option, value)
-
             if save:
                 self.save()
 
@@ -318,7 +309,6 @@ class UserConfig(DefaultsConfig):
         :param option: Config option.
         :returns: Config value or None if section / option do not exist.
         """
-
         with self._lock:
             secdict = self.default_config.get(section, {})
             return secdict.get(option, NoDefault)
@@ -334,9 +324,7 @@ class UserConfig(DefaultsConfig):
         :raises cp.NoSectionError: if the section does not exist.
         :raises cp.NoOptionError: if the option does not exist and no default is given.
         """
-
         with self._lock:
-
             if not self.has_section(section):
                 if default is NoDefault:
                     raise cp.NoSectionError(section)
@@ -378,9 +366,7 @@ class UserConfig(DefaultsConfig):
 
         If the section or option does not exist, it will be created.
         """
-
         with self._lock:
-
             if section not in self.default_config:
                 self.default_config[section] = {}
 
@@ -397,9 +383,7 @@ class UserConfig(DefaultsConfig):
         :param value: Config value.
         :param save: Whether to save the changes to the drive.
         """
-
         with self._lock:
-
             default_value = self.get_default(section, option)
 
             if default_value is NoDefault:
@@ -411,7 +395,7 @@ class UserConfig(DefaultsConfig):
 
             if type(default_value) is not type(value):
                 raise ValueError(
-                    f"Inconsistent config type for [{section}][{option}]. "
+                    f"Inconsistent type for config value [{section}][{option}]. "
                     f"Expected {default_value.__class__.__name__} but "
                     f"got {value.__class__.__name__}."
                 )
@@ -429,7 +413,6 @@ class UserConfig(DefaultsConfig):
         :param save: Whether to save the changes to the drive.
         :returns: Whether the section was removed successfully.
         """
-
         with self._lock:
             res = super().remove_section(section)
             if save:
@@ -445,7 +428,6 @@ class UserConfig(DefaultsConfig):
         :param save: Whether to save the changes to the drive.
         :returns: Whether the section was removed successfully.
         """
-
         with self._lock:
             res = super().remove_option(section, option)
             if save:
@@ -454,9 +436,7 @@ class UserConfig(DefaultsConfig):
 
     def cleanup(self) -> None:
         """Remove files associated with config and reset to defaults."""
-
         with self._lock:
-
             self.reset_to_defaults(save=False)
             backup_path = osp.join(self._dirname, self._backup_folder)
 
@@ -481,8 +461,8 @@ class UserConfig(DefaultsConfig):
 # ======================================================================================
 
 
-class PersistentMutableSet(abc.MutableSet):
-    """Wraps a list in our state file as a MutableSet
+class PersistentMutableSet(MutableSet[_T]):
+    """Wraps a list in our state file as a Mapping
 
     :param conf: UserConfig instance to store the set.
     :param section: Section name in state file.
@@ -496,7 +476,7 @@ class PersistentMutableSet(abc.MutableSet):
         self._conf = conf
         self._lock = RLock()
 
-    def __iter__(self) -> Iterator[Any]:
+    def __iter__(self) -> Iterator[_T]:
         with self._lock:
             return iter(self._conf.get(self.section, self.option))
 
@@ -504,32 +484,32 @@ class PersistentMutableSet(abc.MutableSet):
         with self._lock:
             return entry in self._conf.get(self.section, self.option)
 
-    def __len__(self):
+    def __len__(self) -> int:
         with self._lock:
             return len(self._conf.get(self.section, self.option))
 
-    def add(self, entry: Any) -> None:
+    def add(self, entry: _T) -> None:
         with self._lock:
             state_list = self._conf.get(self.section, self.option)
             state_list = set(state_list)
             state_list.add(entry)
             self._conf.set(self.section, self.option, list(state_list))
 
-    def discard(self, entry: Any) -> None:
+    def discard(self, entry: _T) -> None:
         with self._lock:
             state_list = self._conf.get(self.section, self.option)
             state_list = set(state_list)
             state_list.discard(entry)
             self._conf.set(self.section, self.option, list(state_list))
 
-    def update(self, *others: Any) -> None:
+    def update(self, *others: _T) -> None:
         with self._lock:
             state_list = self._conf.get(self.section, self.option)
             state_list = set(state_list)
             state_list.update(*others)
             self._conf.set(self.section, self.option, list(state_list))
 
-    def difference_update(self, *others: Any) -> None:
+    def difference_update(self, *others: _T) -> None:
         with self._lock:
             state_list = self._conf.get(self.section, self.option)
             state_list = set(state_list)
