@@ -13,14 +13,17 @@ from functools import wraps
 from queue import Empty, Queue
 from threading import Event, RLock, Thread
 from tempfile import TemporaryDirectory
-from typing import Iterator, cast, TypeVar, Callable, Any
+from typing import Iterator, TypeVar, Callable, Generic, Any
+from typing_extensions import ParamSpec, Concatenate
 
 # local imports
 from . import __url__
 from . import notify
-from .client import DropboxClient, API_HOST
+from .client import API_HOST
 from .core import TeamRootInfo, UserRootInfo
-from .config import MaestralConfig, MaestralState
+from .fsevents import Observer, ObserverType
+from .config import MaestralConfig, MaestralState, PersistentMutableSet
+from .config.user import UserConfig
 from .constants import (
     DISCONNECTED,
     CONNECTING,
@@ -29,7 +32,6 @@ from .constants import (
     PAUSED,
     CONNECTED,
 )
-from .models import SyncEvent
 from .exceptions import (
     CancelledError,
     DropboxConnectionError,
@@ -40,11 +42,12 @@ from .exceptions import (
     DropboxServerError,
 )
 from .sync import SyncEngine
-from .fsevents import Observer
 from .logging import scoped_logger
+from .notify import MaestralDesktopNotifier
 from .utils import removeprefix
 from .utils.integration import check_connection, get_inotify_limits
 from .utils.path import move, delete, is_equal_or_child, is_child, normalize
+from .utils.integration import get_ac_state, ACState
 
 
 __all__ = ["SyncManager"]
@@ -52,16 +55,17 @@ __all__ = ["SyncManager"]
 DROPBOX_API_HOSTNAME = "https://" + API_HOST
 
 
-FT = TypeVar("FT", bound=Callable[..., Any])
+P = ParamSpec("P")
+T = TypeVar("T")
 
-malloc_trim: Callable
+malloc_trim: Callable[[int], None]
 
 try:
     libc = ctypes.CDLL("libc.so.6")
     malloc_trim = libc.malloc_trim
 except (OSError, AttributeError):
 
-    def malloc_trim(*args):
+    def malloc_trim(pad: int) -> None:
         pass
 
 
@@ -71,22 +75,62 @@ def _free_memory() -> None:
     malloc_trim(0)
 
 
+class PersistentQueue(Generic[T]):
+    def __init__(self, conf: UserConfig, section: str, option: str) -> None:
+        self._lock = RLock()
+        self._queue: Queue[T] = Queue()
+        self._persistent: PersistentMutableSet[T] = PersistentMutableSet(
+            conf, section, option
+        )
+
+        for item in self._persistent:
+            self._queue.put(item)
+
+    def qsize(self) -> int:
+        return self._queue.qsize()
+
+    def has_pending(self) -> bool:
+        return not self._queue.empty()
+
+    def put(self, item: T) -> None:
+        with self._lock:
+            self._persistent.add(item)
+            self._queue.put(item)
+
+    def get(self, block: bool = True, timeout: int | None = None) -> T:
+        return self._queue.get(block, timeout)
+
+    def join(self) -> None:
+        self._queue.join()
+
+    def task_done(self, item: T) -> None:
+        with self._lock:
+            self._persistent.discard(item)
+            self._queue.task_done()
+
+    def __contains__(self, entry: Any) -> bool:
+        return entry in self._persistent
+
+
 class SyncManager:
     """Class to manage sync threads
 
-    :param client: The Dropbox API client, a wrapper around the Dropbox Python SDK.
+    :param sync: The SyncEngine.
+    :param desktop_notifier: Used to send desktop notifications for management-level
+        events such as joining or leaving a team or fatal errors.
     """
 
-    added_item_queue: Queue[str]
-    """Queue of dropbox paths which have been newly included in syncing."""
+    download_queue: PersistentQueue[str]
+    """Queue of remote paths which have been newly included in syncing."""
 
-    def __init__(self, client: DropboxClient):
-
-        self.client = client
-        self.config_name = self.client.config_name
-        self._conf = MaestralConfig(self.config_name)
-        self._state = MaestralState(self.config_name)
-        self._logger = scoped_logger(__name__, self.config_name)
+    def __init__(
+        self, sync: SyncEngine, desktop_notifier: MaestralDesktopNotifier | None = None
+    ) -> None:
+        self.sync = sync
+        self.desktop_notifier = desktop_notifier
+        self._conf = MaestralConfig(self.sync.config_name)
+        self._state = MaestralState(self.sync.config_name)
+        self._logger = scoped_logger(__name__, self.sync.config_name)
 
         self._lock = RLock()
 
@@ -94,9 +138,7 @@ class SyncManager:
         self.startup_completed = Event()
         self.autostart = Event()
 
-        self.added_item_queue = Queue()
-
-        self.sync = SyncEngine(self.client)
+        self.download_queue = PersistentQueue(self._state, "sync", "pending_downloads")
 
         self._startup_time = -1.0
 
@@ -110,15 +152,17 @@ class SyncManager:
         )
         self.connection_helper.start()
 
-        self.local_observer_thread: Observer | None = None
+        self.local_observer_thread: ObserverType | None = None
 
-    def _with_lock(fn: FT) -> FT:  # type: ignore
+    def _with_lock(  # type:ignore[misc]
+        fn: Callable[Concatenate[SyncManager, P], T]
+    ) -> Callable[Concatenate[SyncManager, P], T]:
         @wraps(fn)
-        def wrapper(self, *args, **kwargs):
-            with self._lock:
-                return fn(self, *args, **kwargs)
+        def wrapper(__self: SyncManager, *args: P.args, **kwargs: P.kwargs) -> T:
+            with __self._lock:
+                return fn(__self, *args, **kwargs)
 
-        return cast(FT, wrapper)
+        return wrapper
 
     # ---- config and state ------------------------------------------------------------
 
@@ -134,25 +178,12 @@ class SyncManager:
         """Setter: reindex_interval"""
         self._conf.set("sync", "reindex_interval", interval)
 
-    @property
-    def activity(self) -> dict[str, SyncEvent]:
-        """Returns a list all items queued for or currently syncing."""
-        return self.sync.syncing
-
-    @property
-    def history(self) -> list[SyncEvent]:
-        """A list of the last SyncEvents in our history. History will be kept for the
-        interval specified by the config value ``keep_history`` (defaults to two weeks)
-        but at most 1,000 events will be kept."""
-        return self.sync.history
-
     @property
     def idle_time(self) -> float:
         """
         Returns the idle time in seconds since the last file change or since startup if
         there haven't been any changes in our current session.
         """
-
         now = time.time()
         time_since_startup = now - self._startup_time
         time_since_last_sync = now - self.sync.last_change
@@ -164,7 +195,6 @@ class SyncManager:
     @_with_lock
     def start(self) -> None:
         """Creates observer threads and starts syncing."""
-
         if self.running.is_set():
             return
 
@@ -190,7 +220,6 @@ class SyncManager:
         )
 
         if self._conf.get("sync", "download"):
-
             self.download_thread = Thread(
                 target=self.download_worker,
                 daemon=True,
@@ -201,7 +230,6 @@ class SyncManager:
                 ),
                 name="maestral-download",
             )
-
             self.download_thread_added_folder = Thread(
                 target=self.download_worker_added_item,
                 daemon=True,
@@ -217,7 +245,6 @@ class SyncManager:
         enable_download = self._conf.get("sync", "download")
 
         if enable_upload:
-
             self.upload_thread = Thread(
                 target=self.upload_worker,
                 daemon=True,
@@ -234,9 +261,10 @@ class SyncManager:
                     self.local_observer_thread = self._create_observer()
                 except MaestralApiError as exc:
                     self._logger.error(exc.title, exc_info=True)
-                    self.sync.desktop_notifier.notify(
-                        exc.title, exc.message, level=notify.ERROR
-                    )
+                    if self.desktop_notifier:
+                        self.desktop_notifier.notify(
+                            exc.title, exc.message, level=notify.ERROR
+                        )
                     return
 
         self.running.set()
@@ -251,11 +279,9 @@ class SyncManager:
             self.download_thread_added_folder.start()
 
         self.startup_thread.start()
-
         self._startup_time = time.time()
 
-    def _create_observer(self) -> Observer:
-
+    def _create_observer(self) -> ObserverType:
         local_observer_thread = Observer(timeout=40)
         local_observer_thread.name = "maestral-fsobserver"
         local_observer_thread.schedule(
@@ -269,9 +295,7 @@ class SyncManager:
         try:
             local_observer_thread.start()
         except OSError as exc:
-
             if exc.errno in (errno.ENOSPC, errno.EMFILE):
-
                 try:
                     max_user_watches, max_user_instances, _ = get_inotify_limits()
                 except OSError:
@@ -324,7 +348,6 @@ class SyncManager:
     @_with_lock
     def stop(self) -> None:
         """Stops syncing and destroys worker threads."""
-
         if self.running.is_set():
             self._logger.info("Shutting down threads...")
 
@@ -343,12 +366,10 @@ class SyncManager:
 
     def reset_sync_state(self) -> None:
         """Resets all saved sync state. Settings are not affected."""
-
         if self.running.is_set():
             raise MaestralApiError(
                 "Cannot reset sync state while syncing", "Please try again when idle."
             )
-
         self.sync.reset_sync_state()
 
     def rebuild_index(self) -> None:
@@ -360,18 +381,30 @@ class SyncManager:
 
         Rebuilding will be performed asynchronously.
         """
-
         self._logger.info("Rebuilding index...")
 
         was_running = self.running.is_set()
 
         self.stop()
-
         self.reset_sync_state()
 
         if was_running:
             self.start()
 
+    def _should_rebuild_index(self) -> bool:
+        """
+        Check if reindexing is due and can be performed now. This is determined by the
+        'reindex_interval' setting. Don't reindex if we are running on battery power.
+        """
+        elapsed = time.time() - self.sync.last_reindex
+        ac_state = get_ac_state()
+
+        reindexing_due = elapsed > self.reindex_interval
+        is_idle = self.idle_time > 20 * 60
+        has_ac_power = ac_state in (ACState.Connected, ACState.Undetermined)
+
+        return reindexing_due and is_idle and has_ac_power
+
     # ---- path root management --------------------------------------------------------
 
     def check_and_update_path_root(self) -> bool:
@@ -382,13 +415,9 @@ class SyncManager:
 
         :returns: Whether the path root was updated.
         """
-
         if self._needs_path_root_update():
-
             was_running = self.running.is_set()
-
             self.stop()
-
             self._update_path_root()
 
             if was_running:
@@ -407,18 +436,16 @@ class SyncManager:
 
         :returns: Whether the configured root namespace needs to be updated.
         """
-
         self._logger.debug("Checking path root...")
 
-        account_info = self.client.get_account_info()
-        return self.client.namespace_id != account_info.root_info.root_namespace_id
+        account_info = self.sync.client.get_account_info()
+        return self.sync.client.namespace_id != account_info.root_info.root_namespace_id
 
     def _update_path_root(self) -> None:
         """
         Changes the layout of the local Dropbox folder if the user joins or leaves a
         team. New team folders will be downloaded, old team folders will be removed.
         """
-
         current_root_type = self._state.get("account", "path_root_type")
         current_user_home_path = self._state.get("account", "home_path")
         current_user_home_path_lower = normalize(current_user_home_path)
@@ -429,8 +456,8 @@ class SyncManager:
                 "Inconsistent namespace information found.",
             )
 
-        root_info = self.client.account_info.root_info
-        team = self.client.account_info.team
+        root_info = self.sync.client.account_info.root_info
+        team = self.sync.client.account_info.team
 
         team_name = team.name if team else "team"
 
@@ -459,23 +486,19 @@ class SyncManager:
             raise NoDropboxDirError(title, msg)
 
         with self.sync.sync_lock:
-
             if new_root_type == "team" and current_root_type == "user":
-
                 # User joined a team.
-
                 self._logger.info("User joined %s. Resyncing user files.", team_name)
-
-                self.sync.desktop_notifier.notify(
-                    f"Joined {team_name}",
-                    "Migrating user files and downloading team folders",
-                )
+                if self.desktop_notifier:
+                    self.desktop_notifier.notify(
+                        f"Joined {team_name}",
+                        "Migrating user files and downloading team folders",
+                    )
 
                 # Migrate user folder to "self.sync.dropbox_path/home_path". We do this
                 # by creating a temporary folder and renaming it after moving all
                 # personal items to prevent name conflicts where a folder has the same
                 # name as `root_info.home_path`.
-
                 tmpdir = TemporaryDirectory(dir=self.sync.dropbox_path)
 
                 for entry in local_dropbox_dirlist:
@@ -497,15 +520,13 @@ class SyncManager:
                 self.sync.excluded_items = new_excluded
 
             elif new_root_type == "user" and current_root_type == "team":
-
                 # User left a team.
-
                 self._logger.info("User left team. Updating folder layout.")
-
-                self.sync.desktop_notifier.notify(
-                    "Left Dropbox Team",
-                    "Migrating user files and removing team folders",
-                )
+                if self.desktop_notifier:
+                    self.desktop_notifier.notify(
+                        "Left Dropbox Team",
+                        "Migrating user files and removing team folders",
+                    )
 
                 # Remove all team folders.
                 for entry in local_dropbox_dirlist:
@@ -515,7 +536,6 @@ class SyncManager:
                 # Migrate user folders to local Dropbox root. We do this by renaming the
                 # user home to a temporary name and then moving its contents to the
                 # parent folder.
-
                 old_home_root = self.sync.dropbox_path + current_user_home_path
 
                 tmpdir = TemporaryDirectory(dir=self.sync.dropbox_path)
@@ -549,14 +569,12 @@ class SyncManager:
                 self.sync.excluded_items = new_excluded
 
             elif new_root_type == "team" and current_root_type == "team":
-
                 # User switched between different teams.
-
                 self._logger.info("User switched teams. Updating team folders.")
-
-                self.sync.desktop_notifier.notify(
-                    f"Switched teams to {team_name}", "Updating team folders"
-                )
+                if self.desktop_notifier:
+                    self.desktop_notifier.notify(
+                        f"Switched teams to {team_name}", "Updating team folders"
+                    )
 
                 # Remove all team folders, leave user folder alone.
                 for entry in local_dropbox_dirlist:
@@ -576,7 +594,7 @@ class SyncManager:
                 self.sync.excluded_items = new_excluded
 
             # Update path root of client.
-            self.client.update_path_root(root_info)
+            self.sync.client.update_path_root(root_info)
 
             #  Trigger reindex.
             self.sync.reset_sync_state()
@@ -589,9 +607,7 @@ class SyncManager:
         is lost and resumes syncing when reconnected and syncing has not been paused by
         the user.
         """
-
         while self._connection_helper_running:
-
             connected = check_connection(DROPBOX_API_HOSTNAME)
 
             if connected != self.connected:
@@ -619,35 +635,32 @@ class SyncManager:
         :param startup_completed: Set when startup sync is completed.
         :param autostart: Set when syncing should automatically resume on connection.
         """
-
         startup_completed.wait()
 
         while running.is_set():
-
             with self._handle_sync_thread_errors(running, autostart):
+                has_changes = self.sync.wait_for_remote_changes(self.sync.remote_cursor)
 
-                with self.sync.client.clone_with_new_session() as client:
+                # Check for root namespace updates. Don't apply any remote
+                # changes in case of a changed root path.
+                if self.check_and_update_path_root():
+                    return
 
-                    has_changes = self.sync.wait_for_remote_changes(
-                        self.sync.remote_cursor, client=client
-                    )
+                # Check for and perform reindexing.
+                if self._should_rebuild_index():
+                    self.rebuild_index()
 
-                    # Check for root namespace updates. Don't apply any remote
-                    # changes in case of a changed root path.
-                    if self.check_and_update_path_root():
-                        return
+                if not running.is_set():
+                    return
 
-                    if not running.is_set():
-                        return
+                self.sync.ensure_dropbox_folder_present()
 
-                    self.sync.ensure_dropbox_folder_present()
+                if has_changes:
+                    self._logger.info(SYNCING)
+                    self.sync.download_sync_cycle()
+                    self._logger.info(IDLE)
 
-                    if has_changes:
-                        self._logger.info(SYNCING)
-                        self.sync.download_sync_cycle(client)
-                        self._logger.info(IDLE)
-
-                        client.get_space_usage()
+                    self.sync.client.get_space_usage()
 
         _free_memory()
 
@@ -664,32 +677,23 @@ class SyncManager:
         :param startup_completed: Set when startup sync is completed.
         :param autostart: Set when syncing should automatically resume on connection.
         """
-
         startup_completed.wait()
 
         while running.is_set():
-
             with self._handle_sync_thread_errors(running, autostart):
-
                 try:
-                    dbx_path_lower = self.added_item_queue.get(timeout=40)
+                    dbx_path_lower = self.download_queue.get(timeout=40)
                 except Empty:
-                    pass
-                else:
-                    # Guard against crashes.
-                    self.sync.pending_downloads.add(dbx_path_lower)
+                    continue
 
-                    if not running.is_set():
-                        return
+                if not running.is_set():
+                    self.download_queue.put(dbx_path_lower)
+                    return
 
-                    with self.sync.sync_lock:
-
-                        with self.sync.client.clone_with_new_session() as client:
-                            self.sync.get_remote_item(dbx_path_lower, client)
-
-                        self.sync.pending_downloads.discard(dbx_path_lower)
-
-                        self._logger.info(IDLE)
+                with self.sync.sync_lock:
+                    self.sync.get_remote_item(dbx_path_lower)
+                    self.download_queue.task_done(dbx_path_lower)
+                    self._logger.info(IDLE)
 
         _free_memory()
 
@@ -706,13 +710,10 @@ class SyncManager:
         :param startup_completed: Set when startup sync is completed.
         :param autostart: Set when syncing should automatically resume on connection.
         """
-
         startup_completed.wait()
 
         while running.is_set():
-
             with self._handle_sync_thread_errors(running, autostart):
-
                 has_changes = self.sync.wait_for_local_changes()
 
                 if not running.is_set():
@@ -740,16 +741,14 @@ class SyncManager:
         :param startup_completed: Set when startup sync is completed.
         :param autostart: Set when syncing should automatically resume on connection.
         """
-
         with self._handle_sync_thread_errors(running, autostart):
-
             # Fail early if Dropbox folder disappeared.
             self.sync.ensure_dropbox_folder_present()
 
             # Reload mignore rules.
             self.sync.load_mignore_file()
 
-            self.client.get_space_usage()
+            self.sync.client.get_space_usage()
 
             # Update path root and migrate local folders. This is required when a user
             # joins or leaves a team and their root namespace changes.
@@ -759,28 +758,27 @@ class SyncManager:
                 startup_completed.set()
                 return
 
-            with self.sync.client.clone_with_new_session() as client:
+            # Retry failed downloads.
+            if len(self.sync.download_errors) > 0:
+                self._logger.info("Retrying failed syncs...")
 
-                # Retry failed downloads.
-                if len(self.sync.download_errors) > 0:
-                    self._logger.info("Retrying failed syncs...")
+            for error in list(self.sync.download_errors):
+                self.sync.get_remote_item(error.dbx_path_lower)
 
-                for error in list(self.sync.download_errors):
-                    self.sync.get_remote_item(error.dbx_path_lower, client)
+            # Resume interrupted downloads.
+            if self.download_queue.qsize() > 0:
+                self._logger.info("Resuming interrupted syncs...")
 
-                # Resume interrupted downloads.
-                if len(self.sync.pending_downloads) > 0:
-                    self._logger.info("Resuming interrupted syncs...")
+            while self.download_queue.has_pending():
+                dbx_path = self.download_queue.get()
+                self.sync.get_remote_item(dbx_path)
+                self.download_queue.task_done(dbx_path)
 
-                for dbx_path in list(self.sync.pending_downloads):
-                    self.sync.get_remote_item(dbx_path, client)
-                    self.sync.pending_downloads.discard(dbx_path)
+            if not running.is_set():
+                startup_completed.set()
+                return
 
-                if not running.is_set():
-                    startup_completed.set()
-                    return
-
-                self.sync.download_sync_cycle(client)
+            self.sync.download_sync_cycle()
 
             if not running.is_set():
                 startup_completed.set()
@@ -800,7 +798,6 @@ class SyncManager:
     def _handle_sync_thread_errors(
         self, running: Event, autostart: Event
     ) -> Iterator[None]:
-
         try:
             yield
         except CancelledError:
@@ -819,10 +816,11 @@ class SyncManager:
             title = getattr(err, "title", "Unexpected error")
             message = getattr(err, "message", "Please restart to continue syncing")
             self._logger.error(title, exc_info=True)
-            self.sync.desktop_notifier.notify(title, message, level=notify.ERROR)
+            if self.desktop_notifier:
+                self.desktop_notifier.notify(title, message, level=notify.ERROR)
             self.stop()
 
-    def __del__(self):
+    def __del__(self) -> None:
         try:
             self.stop()
             self._connection_helper_running = False
