@@ -1,46 +1,49 @@
-# -*- coding: utf-8 -*-
 """
 This module provides functions for platform integration. Most of the functionality here
-could also be achieved with psutils but we want to avoid the large dependency.
+could also be achieved with psutils, but we want to avoid the large dependency.
 """
+from __future__ import annotations
 
 import os
 import platform
 import enum
 import resource
-import socket
+import requests
 import time
+import logging
+import socket
 from pathlib import Path
-from typing import Union, Tuple
+from typing import Union, Tuple, Optional
+from urllib.parse import urlparse
 
 __all__ = [
+    "cat",
     "get_ac_state",
     "ACState",
     "get_inotify_limits",
-    "CPU_COUNT",
+    "CPU_CORE_COUNT",
     "cpu_usage_percent",
     "check_connection",
+    "SystemdNotifier",
 ]
 
 
-CPU_COUNT = os.cpu_count() or 1  # os.cpu_count can return None
+CPU_CORE_COUNT = os.cpu_count() or 1  # os.cpu_count can return None
 LINUX_POWER_SUPPLY_PATH = "/sys/class/power_supply"
 
 
-def multi_cat(*paths: Path) -> Union[int, bytes, None]:
+def cat(*paths: "os.PathLike[str]") -> Union[bytes, None]:
     """
     Attempts to read the content of multiple files which may not exist. Returns the
     content of the first file which can be read. If none of them can be read return
     None. Returns an integer if the content is a digit.
     """
-
     for path in paths:
         try:
-            ret = path.read_bytes().strip()
+            with open(path, "rb") as f:
+                return f.read().strip()
         except OSError:
             pass
-        else:
-            return int(ret) if ret.isdigit() else ret
 
     return None
 
@@ -59,9 +62,7 @@ def get_ac_state() -> ACState:
 
     :returns: ``True`` if the device has AC power, ``False`` otherwise.
     """
-
     if platform.system() == "Darwin":
-
         from ctypes import c_double
         from rubicon.objc.runtime import load_library
 
@@ -78,37 +79,35 @@ def get_ac_state() -> ACState:
             return ACState.Disconnected
 
     elif platform.system() == "Linux":
-
         # taken from https://github.com/giampaolo/psutil
 
-        supplies = list(os.scandir(LINUX_POWER_SUPPLY_PATH))
+        supply_entry = list(os.scandir(LINUX_POWER_SUPPLY_PATH))
 
         ac_paths = [
-            Path(s.path)
-            for s in supplies
-            if s.name.startswith("A") or "ac" in s.name.lower()
+            Path(entry.path)
+            for entry in supply_entry
+            if entry.name.startswith("A") or "ac" in entry.name.lower()
         ]
 
-        bat_paths = [
-            Path(s.path)
-            for s in supplies
-            if s.name.startswith("B") or "battery" in s.name.lower()
+        battery_paths = [
+            Path(entry.path)
+            for entry in supply_entry
+            if entry.name.startswith("B") or "battery" in entry.name.lower()
         ]
 
-        online = multi_cat(*iter(path / "online" for path in ac_paths))
+        online = cat(*iter(path / "online" for path in ac_paths))
 
         if online is not None:
-            if online == 1:
+            if online == b"1":
                 return ACState.Connected
             else:
                 return ACState.Disconnected
 
-        elif len(bat_paths) > 0:
-
+        elif len(battery_paths) > 0:
             # Get the first available battery. Usually this is "BAT0", except
             # some rare exceptions:
             # https://github.com/giampaolo/psutil/issues/1238
-            bat0 = sorted(bat_paths)[0]
+            bat0 = sorted(battery_paths)[0]
 
             try:
                 status = (bat0 / "status").read_text().strip().lower()
@@ -131,7 +130,6 @@ def get_inotify_limits() -> Tuple[int, int, int]:
     :raises OSError: if the settings cannot be read from /proc/sys/fs/inotify. This may
         happen if /proc/sys is left out of the kernel image or simply not mounted.
     """
-
     root = Path("/proc/sys/fs/inotify")
 
     max_user_watches_path = root / "max_user_watches"
@@ -163,12 +161,11 @@ def cpu_usage_percent(interval: float = 0.1) -> float:
     :param interval: Interval in sec between comparisons of CPU times.
     :returns: CPU usage during interval in percent.
     """
-
     if interval <= 0:
         raise ValueError(f"interval is not positive (got {interval!r})")
 
-    def timer():
-        return time.monotonic() * CPU_COUNT
+    def timer() -> float:
+        return time.monotonic() * CPU_CORE_COUNT
 
     st1 = timer()
     rt1 = resource.getrusage(resource.RUSAGE_SELF)
@@ -184,22 +181,68 @@ def cpu_usage_percent(interval: float = 0.1) -> float:
     except ZeroDivisionError:
         return 0.0
     else:
-        single_cpu_percent = overall_cpus_percent * CPU_COUNT
+        single_cpu_percent = overall_cpus_percent * CPU_CORE_COUNT
         return round(single_cpu_percent, 1)
 
 
-def check_connection(hostname: str, timeout: int = 2) -> bool:
+def check_connection(
+    hostname: str, timeout: int = 2, logger: Optional[logging.Logger] = None
+) -> bool:
     """
     A low latency check for an internet connection.
 
     :param hostname: Hostname to use for connection check.
     :param timeout: Timeout in seconds for connection check.
+    :param logger: If provided, log output for connection failures will be logged to
+        this logger with the level DEBUG.
     :returns: Connection availability.
     """
+    if urlparse(hostname).scheme not in ["http", "https"]:
+        hostname = "http://" + hostname
     try:
-        host = socket.gethostbyname(hostname)
-        s = socket.create_connection(address=(host, 80), timeout=timeout)
-        s.close()
+        requests.head(hostname, timeout=timeout)
         return True
     except Exception:
+        if logger:
+            logger.debug("Could not reach %s", hostname, exc_info=True)
         return False
+
+
+class SystemdNotifier:
+    """
+    An interface to notify the systemd the service manager about status changes.
+
+    Sends a status message to the systemd the service manager on the socket address
+    provided by the NOTIFY_SOCKET environment variable. Does nothing NOTIFY_SOCKET is
+    not set.
+
+    See https://www.freedesktop.org/software/systemd/man/sd_notify.html for a
+    documentation of message formats expected by systemd.
+    """
+
+    def __init__(self) -> None:
+        self._socket = None
+
+        addr = os.getenv("NOTIFY_SOCKET")
+        if addr is None:
+            return
+        elif addr[0] == "@":
+            addr = "\0" + addr[1:]
+
+        try:
+            self._socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
+            self._socket.connect(addr)
+        except OSError:
+            pass
+
+    def notify(self, status: str) -> None:
+        """
+        Send a status update to the service manager.
+
+        :param status: The status update to send.
+        """
+        if self._socket:
+            try:
+                self._socket.sendall(status.encode(errors="replace"))
+            except OSError:
+                pass
